import logging
import os

from satella.exceptions import ProcessFailed
from satella.processes import call_and_return_stdout
import sys
import typing as tp

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.DEBUG)


# unceremoniously taken from
# https://codereview.stackexchange.com/questions/190886/platform-independent-whereis-function
# and modified to better suit UNIX platforms
def whereis(app: str) -> tp.List[str]:
    result = None
    try:
        args = ['whereis', app]
        result = call_and_return_stdout(args,
                                        expected_return_code=0,
                                        encoding='utf-8')
    except ProcessFailed:
        print('whereis not found, aborting')
        sys.exit(1)

    if result is None:
        return []
    else:
        result = result.replace("\r", "").split("\n")
        results = []
        for line in result:
            line = line.strip()
            if not line:
                continue

            if ':' in line:
                results = line.split(':', 1)[1].strip().split(' ')
        return results


def filter_whereis(app: str, abort_on_failure=True) -> tp.Iterator[str]:
    """
    Return a position of an executable. Verify that it's executable.
    """
    found = False
    for path in whereis(app):
        if os.path.isdir(path):
            continue
        if os.access(path, os.X_OK):
            found = True
            yield path

    if not found:
        if not abort_on_failure:
            return
        else:
            print('%s not found, aborting' % (app,))
            sys.exit(1)



#
# Take care, in importing this module sys.argv gets changed!
#
import os
import shutil
import sys

import pkg_resources
from satella.coding import silence_excs
from satella.files import read_in_file, write_to_file

from interceptor.config import load_config_for, Configuration
from interceptor.whereis import filter_whereis

INTERCEPTED = '-intercepted'
INTERCEPTOR_WRAPPER_STRING = 'from interceptor.config import load_config_for'

FORCE = '--force' in sys.argv
if FORCE:
    sys.argv.remove('--force')


def is_intercepted(path_name: str, print_messages=False) -> bool:
    a = False
    with silence_excs(UnicodeDecodeError), open(path_name, 'rb') as f_in:
        intercepted_real = f_in.read(512).decode('utf-8')
        a = INTERCEPTOR_WRAPPER_STRING in intercepted_real
    file_name = os.path.split(path_name)
    if a:
        b = os.path.exists(path_name + INTERCEPTED)
        if not b and print_messages:
            print('%s is intercepted, but %s-intercepted does not exist' % (file_name, file_name))

            if not os.path.exists(os.path.join('/etc/interceptor.d', file_name)):
                print('Additionally, it\'s configuration is not present')

        return True
    else:
        if print_messages:
            print('%s does not contain interceptor code' % (path_name, ))
        return False


def is_all_intercepted(name: str) -> bool:
    for path in filter_whereis(name):
        if not is_intercepted(path):
            return False
    return True


def assert_intercepted(name: str) -> None:
    if FORCE:
        print('Skipping a check to see if %s is intercepted due to --force' % (name, ))
        return
    if is_all_intercepted(name):
        return
    if is_partially_intercepted(name):
        print('''%s is partially intercepted. This means that there exist binaries of %s
'that have not been intercepted. To fix that, call:

intercept %s --force
''' % (name, name, name))
        return
    print('%s is not intercepted' % (name,))
    abort()


def is_partially_intercepted(name: str, print_messages=False) -> bool:
    interceptions = []
    for path in filter_whereis(name):
        if is_intercepted(path, print_messages=print_messages):
            if print_messages:
                print('%s is currently intercepted' % (path,))
            interceptions.append(True)
        else:
            interceptions.append(False)
    return not all(interceptions) and any(interceptions)


def is_completely_unintercepted(name: str) -> bool:
    return not is_all_intercepted(name) and not is_partially_intercepted(name)


def can_be_unintercepted(name: str) -> bool:
    for path in filter_whereis(name):
        if not is_intercepted(path):
            print('%s is not intercepted' % (path,))
            return False
        target_path = path + INTERCEPTED
        if not os.path.exists(target_path):
            print('%s does not exist' % (path,))
            return False
    return True


def abort():
    print('Aborting.')
    sys.exit(1)


def unintercept_path(path_name: str) -> None:
    src_name = path_name + INTERCEPTED
    shutil.move(src_name, path_name)
    print('Successfully unintercepted %s' % (path_name,))


def intercept_path(tool_name: str, file_name: str) -> None:
    source_file = pkg_resources.resource_filename(__name__, 'templates/cmdline.py')
    target_intercepted = file_name + INTERCEPTED
    previous_chmod = os.stat(file_name).st_mode & 0o777
    shutil.copy(file_name, target_intercepted)
    os.unlink(file_name)
    source_content = read_in_file(source_file, 'utf-8')
    source_content = source_content.format(EXECUTABLE=sys.executable,
                                           TOOLNAME=tool_name,
                                           LOCATION=target_intercepted,
                                           VERSION=pkg_resources.require('interceptor')[0].version)
    write_to_file(file_name, source_content, 'utf-8')
    os.chmod(file_name, previous_chmod)
    print('Successfully intercepted %s' % (file_name,))


def intercept_tool(tool_name: str):
    if is_partially_intercepted(tool_name):
        if not FORCE:
            print(
                '%s is partially intercepted. Use --force if you want to continue.' % (tool_name,))
            abort()

    if is_all_intercepted(tool_name):
        print('%s is completely intercepted.' % (tool_name,))
        abort()

    for path in filter_whereis(tool_name):
        if not is_intercepted(path):
            intercept_path(tool_name, path)

    try:
        load_config_for(tool_name, None)
        print('Config for %s already exists' % (tool_name,))
    except KeyError:
        print('Config for %s not found, creating a fresh one' % (tool_name,))
        Configuration(app_name=tool_name).save()
    except ValueError:
        print('Config for %s exists, but is invalid. Usage of %s will be impossible until '
              'this is fixed' % (tool_name, tool_name))


def unintercept_tool(tool_name: str):
    if not can_be_unintercepted(tool_name):
        if not FORCE:
            print('%s cannot be unintercepted. Use --force to proceed' % (tool_name,))
            abort()

    for path in filter_whereis(tool_name):
        if is_intercepted(path):
            unintercept_path(path)
        else:
            print('Skipping on %s' % (path,))
    print('Unintercepted %s, leaving the configuration in-place' % (tool_name,))


def check(tool_name: str, add_config: bool = False):
    total_interception = is_all_intercepted(tool_name)
    partial_interception = is_partially_intercepted(tool_name, True)
    if not total_interception and not partial_interception:
        print('%s is not intercepted at all' % (tool_name,))
        sys.exit(0)

    if partial_interception:
        print('''%s is partially intercepted. To clean this up, call:
intercept %s --force
''' % (tool_name, tool_name))

    cfg_exists = False
    try:
        cfg = load_config_for(tool_name, None)
        cfg_exists = True
        print('Configuration for %s exists and is valid' % (tool_name,))
    except ValueError as e:
        print('Configuration for %s is invalid JSON.\nDetails: %s' % (tool_name, e.args[0]))
    except KeyError:
        if add_config:
            print('%s configuration not found, creating a new one' % (tool_name,))
            cfg = Configuration(app_name=tool_name)
            cfg_exists = True
    if cfg_exists:
        if os.path.islink(cfg.path):
            target = os.readlink(cfg.path).split('/')[-1]
            print('%s config is a symlink to %s config' % (tool_name, target))
        cfg.save()


def link(app_name, target_name, copy=False):
    assert_intercepted(app_name)
    assert_intercepted(target_name)
    source = os.path.join('/etc/interceptor.d', app_name)
    target = os.path.join('/etc/interceptor.d', target_name)
    if os.path.islink(source) and not FORCE and not copy:
        print('Refusing to link, since %s is already a symlink!' % (app_name,))
        abort()
    with silence_excs(IOError):
        os.unlink(target)
    if copy:
        shutil.copy(source, target)
    else:
        os.symlink(source, target)
    if not copy:
        print('Linked %s to read from %s\'s config' % (target_name, app_name))
    else:
        print('Copied %s to read from %s\'s config' % (target_name, app_name))


def assert_etc_interceptor_d_exists():
    if not os.path.exists('/etc/interceptor.d'):
        print('/etc/interceptor.d does not exist, creating...')
        os.mkdir('/etc/interceptor.d')


def edit(app_name):
    assert_intercepted(app_name)
    editor = list(filter_whereis('nano', abort_on_failure=False))
    if not editor:
        editor = list(filter_whereis('vi', abort_on_failure=False))
        if not editor:
            print('Neither nano nor vi were found')
            abort()
    os.execv(editor[0], [editor[0], os.path.join('/etc/interceptor.d', app_name)])


def reset(app_name):
    assert_intercepted(app_name)
    path = os.path.join('/etc/interceptor.d', app_name)
    if os.path.islink(path):
        target = os.readlink(path).split('/')[-1]
        print('%s config was previously a symlink to %s' % (app_name, target))
    os.unlink(path)
    Configuration(app_name=app_name).save()
    print('Configuration for %s reset' % (app_name, ))


def configure(op_name, app_name, target_name):
    assert_intercepted(app_name)
    cfg = load_config_for(app_name, None)
    if op_name == 'append':
        cfg.arg_names_to_append.append(target_name)
    elif op_name == 'prepend':
        cfg.arg_names_to_prepend.append(target_name)
    elif op_name == 'disable':
        cfg.arg_names_to_disable.append(target_name)
    elif op_name == 'replace':
        cfg.arg_names_to_replace.append([target_name, sys.target_namev[4]])
    elif op_name == 'display':
        cfg.display_before_start = True
    elif op_name == 'hide':
        cfg.display_before_start = False
    elif op_name == 'notify':
        cfg.notify_about_actions = True
    elif op_name == 'unnotify':
        cfg.notify_about_actions = False
    cfg.save()
    print('Configuration changed')



#!{EXECUTABLE}

# Generated automatically by interceptor, a tool to intercept calls
# to the commands and to alter their arguments.

# To learn more visit https://github.com/Cervi-Robotics/interceptor

import os
import sys
from interceptor.config import load_config_for

TOOLNAME = '{TOOLNAME}'
LOCATION = '{LOCATION}'
VERSION = '{VERSION}'

if __name__ == '__main__':
    cfg = load_config_for(TOOLNAME, VERSION)
    args = cfg.modify(sys.argv)
    os.execv(LOCATION, args)



import json
import os
import sys

from satella.files import write_to_file, read_in_file

from interceptor.intercepting import intercept_tool, unintercept_tool, assert_intercepted, check, \
    abort, link, assert_etc_interceptor_d_exists, edit, reset, configure


def banner():
    print('''Usage:
    * intercept foo - intercept foo
    * intercept undo foo - cancel intercepting foo
    * intercept configure foo - type in the configuration for foo in JSON format, end with Ctrl+D
    * intercept show foo - show the configuration for foo
    * intercept status foo - display foo's status of interception and details about it's configuration
    * intercept display foo - enable displaying what is launched on foo's startup
    * intercept hide foo - disable displaying what is launched on foo's startup
    * intercept edit foo - launch a nano/vi to edit it's configuration
    * intercept append foo ARG - add ARG to be appended to command line whenever foo is ran
    * intercept prepend foo ARG - add ARG to be prepended to command line whenever foo is ran
    * intercept disable foo ARG - add ARG to be eliminated from the command line whenever foo is ran
    * intercept replace foo ARG1 ARG2 - add ARG1 to be replaced with ARG2 whenever it is passed to foo
    * intercept notify foo - display a notification each time an argument action is taken
    * intercept unnotify foo - hide the notification each time an argument action is taken
    * intercept link foo bar - symlink bar's config file to that of foo
    * intercept copy foo bar - copy foo's configuration onto that of bar
    * intercept reset foo - reset foo's configuration (delete it and create a new one)
    
Use the optional switch --force is you need a command to complete despite the command telling you 
that it is impossible to complete. One trick: already intercepted files won't be intercepted, because
that would lead to overwriting of the original executable, so interceptor won't do that.
''')


def run():
    assert_etc_interceptor_d_exists()

    if len(sys.argv) == 2:
        intercept_tool(sys.argv[1])
    elif len(sys.argv) >= 3:
        op_name = sys.argv[1]
        app_name = sys.argv[2]
        target_name = sys.argv[3] if len(sys.argv) >= 4 else None

        if op_name == 'undo':
            unintercept_tool(app_name)
        elif op_name == 'configure':
            assert_intercepted(app_name)
            data = sys.stdin.read()
            try:
                json.loads(data)
            except json.JSONDecoder:
                print('Configuration is invalid JSON')
                abort()
            write_to_file(os.path.join('/etc/interceptor.d', app_name), data, 'utf-8')
            print('Configuration successfully written')
        elif op_name == 'show':
            assert_intercepted(app_name)
            config = read_in_file(os.path.join('/etc/interceptor.d', app_name), 'utf-8')
            print(config)
        elif op_name == 'status':
            check(app_name, add_config=False)
        elif op_name == 'edit':
            edit(app_name)
        elif op_name in ('append', 'prepend', 'disable', 'replace', 'display', 'hide',
                         'notify', 'unnotify'):
            configure(op_name, app_name, target_name)
        elif op_name == 'link':
            link(app_name, target_name)
        elif op_name == 'copy':
            link(app_name, target_name, copy=True)
        elif op_name == 'reset':
            reset(app_name)
        else:
            print('Unrecognized command %s' % (op_name,))
            banner()
            sys.exit(1)
    else:
        banner()



from setuptools import setup

setup(
    name='cmd-interceptor',
    version='2.0',
    author='Piotr Maślanka',
    install_requires=['satella'],
    package_data={'interceptor': ['templates/cmdline.py',
                                  'templates/config']},
    packages=[
        'interceptor',
    ],
    entry_points={
        'console_scripts': [
            'intercept = interceptor.run:run'
        ]
    },
)



import copy
import os
import sys
import typing as tp
import warnings

import pkg_resources
from satella.coding import for_argument
from satella.json import read_json_from_file, write_json_to_file


class Configuration:
    @property
    def path(self) -> str:
        return os.path.join('/etc/interceptor.d', self.app_name)

    def __init__(self, args_to_disable: tp.Optional[tp.List[str]] = None,
                 args_to_append: tp.Optional[tp.List[str]] = None,
                 args_to_prepend: tp.Optional[tp.List[str]] = None,
                 args_to_replace: tp.Optional[tp.List[tp.Tuple[str, str]]] = None,
                 display_before_start: bool = False,
                 notify_about_actions: bool = False,
                 app_name: tp.Optional[str] = None):
        self.args_to_disable = args_to_disable or []
        self.args_to_append = args_to_append or []
        self.args_to_prepend = args_to_prepend or []
        self.args_to_replace = args_to_replace or []
        self.display_before_start = display_before_start
        self.notify_about_actions = notify_about_actions
        self.app_name = app_name

    def to_json(self):
        return {'args_to_disable': self.args_to_disable,
                'args_to_append': self.args_to_append,
                'args_to_prepend': self.args_to_prepend,
                'args_to_replace': self.args_to_replace,
                'display_before_start': self.display_before_start,
                'notify_about_actions': self.notify_about_actions}

    @for_argument(None, copy.copy)
    def modify(self, args, *extra_args):
        process, *arguments = args
        for arg_to_take_away in self.args_to_disable:
            if arg_to_take_away in arguments:
                if self.notify_about_actions:
                    print('interceptor(%s): taking away %s' % (self.app_name, arg_to_take_away))
                del arguments[arguments.index(arg_to_take_away)]

        for arg_to_append in self.args_to_append:
            if arg_to_append not in arguments:
                if self.notify_about_actions:
                    print('interceptor(%s): appending %s' % (self.app_name, arg_to_append))
                arguments.append(arg_to_append)

        for arg_to_prepend in reversed(self.args_to_prepend):
            if arg_to_prepend not in arguments:
                if self.notify_about_actions:
                    print('interceptor(%s): prepending %s' % (self.app_name, arg_to_prepend))
                arguments = [arg_to_prepend] + arguments

        for arg_to_replace, arg_to_replace_with in self.args_to_replace:
            if arg_to_replace in arguments:
                if self.notify_about_actions:
                    print('interceptor(%s): replacing %s with %s' % (self.app_name, arg_to_replace,
                                                                     arg_to_replace_with))
                arguments[arguments.index(arg_to_replace)] = arg_to_replace_with

        if self.display_before_start:
            print('%s %s' % (sys.argv[0], ' '.join(arguments)))

        return [process, *arguments]

    def save(self):
        write_json_to_file(self.path, self.to_json(), sort_keys=True, indent=4)

    @classmethod
    def from_json(cls, dct, app_name: str):
        prepend = dct.get('args_to_prepend')
        if prepend is None:
            prepend = dct.get('args_to_append_before')
            if prepend is not None:
                warnings.warn('args_to_append_before is deprecated, use args_to_prepend',
                              DeprecationWarning)
        take_away = dct.get('args_to_disable')
        if take_away is None:
            take_away = dct.get('args_to_take_away')
            if take_away is not None:
                warnings.warn('args_to_take_away is deprecated, use args_to_prepend',
                              DeprecationWarning)

        return Configuration(take_away,
                             dct.get('args_to_append'),
                             prepend,
                             dct.get('args_to_replace'),
                             dct.get('display_before_start', False),
                             dct.get('notify_about_actions', False),
                             app_name=app_name)


def assert_correct_version(version: str) -> None:
    if version == '':
        # print('You have used an older version of interceptor to intercept this command.\n'
        #       'It is advised to undo the interception and reintercept the call to upgrade.')
        return
    my_version = pkg_resources.require('interceptor')[0].version
    if int(version.split('.')[0]) > int(my_version.split('.')[0]):
        sys.stderr.write('You have intercepted this call using a higher version of Interceptor. \n'
                         'This might not work as advertised. Try undo\'ing the interception \n'
                         'and intercepting this again.\n'
                         'Aborting.')
        sys.exit(1)


def load_config_for(name: str, version: tp.Optional[str] = '') -> Configuration:
    if version is not None:
        assert_correct_version(version)

    file_name = os.path.join('/etc/interceptor.d', name)
    if not os.path.isfile(file_name):
        print('Configuration for %s does not exist or is not a file' % (name,))
        sys.exit(1)

    cfg = Configuration.from_json(read_json_from_file(file_name), app_name=name)
    return cfg
